#version 330
#extension GL_ARB_explicit_uniform_location : enable
#pragma optionNV(unroll all)

//to deactivate light set color to 0,0,0

#define SPT_LIGHTCOUNT 3  //light count per type
#define DIR_LIGHTCOUNT 2
#define PT_LIGHTCOUNT 6

#define VSM_MIN_VARIANCE 0.0000055
#define VSM_LIGHT_BLEEDING_REDUCTION 0.05 //probably 1.0-x for the values used in literature //see http://http.developer.nvidia.com/GPUGems3/gpugems3_ch08.html light bleeding part... too high will reduce shadow detail, too low will let light bleeding through



//Material:
struct Material
{
	float   lightFactors[5]; //ambientFactor, diffuseFactor, transparency, specularFactor, shininess  
	vec3	ambientColor; 				
	vec3 	diffuseColor; 				

	//texture
	bool	  hasAmbientTex;            
	sampler2D ambientTex;
	vec2 	  ambientTextureScaleUV;
	
	bool	  hasDiffuseTex;
	sampler2D diffuseTex;
	vec2 	  diffuseTextureScaleUV;
	
	float	  bumpFactor;    //if no bump map this is also set to 0
	bool	  bumpMapIsNormalMap; //-> normal map
	sampler2D bumpMap;
	vec2      bumpTextureScaleUV;

	vec3	  specularColor;
	bool      hasSpecularTex;
	sampler2D specularTex;
	vec2      specularTextureScaleUV;
};

layout( location = 20 ) uniform Material mat;



//Lights:
struct DirectionalLight
{
	vec3  color;
	vec4  direction; //standart direciton is 0.0, 0.0, -1.0,... than it gets rotated
	float intensity;
	sampler2D shadowMap;

	//just used in vertex shader:
	mat4  orthoProjMat;
	mat4  transformMat;
	float zFarPlane;
};

struct PointLight
{
	vec3  color;
	vec4  position; //has position & rotation
	float intensity;
	int   decayType;

	//just used in vertex shader:
	mat4  persProjMat;
	mat4  transformMat;
	float zFarPlane;
};

struct SpotLight
{
	vec3  color;
	vec4  position;
	vec4  direction; //standart direciton is 0.0, 0.0, -1.0,... than it gets rotated
	float intensity;
	float innerConeAngle;
	float outerConeAngle;
	int   decayType;
	sampler2D shadowMap;

	//just used in vertex shader
	mat4  persProjMat;
	mat4  transformMat;
	float zFarPlane;
};


//UNIFORMS:
uniform float debugLightFactor;

uniform mat4 viewMatrix;
uniform mat4 projectionMatrix;

uniform DirectionalLight 	dirLights[ DIR_LIGHTCOUNT ];
uniform PointLight			ptLights[   PT_LIGHTCOUNT ];
uniform SpotLight			sptLights[ SPT_LIGHTCOUNT ];

uniform bool				useShadowMapping;


//INs:
flat in vec4 				camWorldPos;
in vec4 					fragmentNormal;
in vec4 					fragment;
in vec2 					fragmentTexCoords;



//bump mapping -> transformation matrix for transforming the normal vector out of the bump map into the view space:
in mat3 tangentToViewSpaceMat;

//calculate the lights in view space:
in vec4 sptLightViewSpacePos[ SPT_LIGHTCOUNT ];
in vec4 sptLightViewSpaceDir[ SPT_LIGHTCOUNT ];
in vec4 ptLightViewSpacePos[  PT_LIGHTCOUNT  ];
in vec4 dirLightViewSpaceDir[ DIR_LIGHTCOUNT ];

layout( location = 50 ) uniform int spotLightsActive;
layout( location = 51 ) uniform int dirLightsActive;
layout( location = 52 ) uniform int pointLightsActive;


//shadow mapping
in vec4 fragmentLightSpt[ SPT_LIGHTCOUNT ];
in float fragmentLightLinearDepthSpt[ SPT_LIGHTCOUNT ];

//in vec4 fragmentLightDir[ DIR_LIGHTCOUNT ];
//in float fragmentLightLinearDepthDir[ DIR_LIGHTCOUNT ]; //linear depth

//not used yet
//in vec4 fragmentLightPt[ PT_LIGHTCOUNT ];
//in float fragmentLightLinearDepthPt[ PT_LIGHTCOUNT ]; //linear depth


//VARIABLES
vec4 N;
vec4 V; //surface position in eye space
vec4 VC; //vector surface camera normalized

//OUT
layout(location = 0, index = 0) out vec4 fragColor;



float linstep(float low, float high, float v){
    return clamp( (v - low)/(high - low), 0.0, 1.0);
}




float calcIntensityAfterDecay( vec4 lightpos, int type, float intensity ){
	if( type == 0 ) return intensity;
	else if( type == 1 ){ //linear
		return intensity / distance( V, lightpos );
	} else if( type == 2 ){ //square
		return intensity / pow( distance( V, lightpos ), 2);
	} 
	//cube
	return intensity / pow( distance( V, lightpos ), 3);
}




float shadowVSM( vec2 moments, float fragLightSpaceDepthLinear ){
	float t = fragLightSpaceDepthLinear;

	float p = float( t <= moments.x );

	float variance = moments.y - (moments.x * moments.x);
	variance = max( VSM_MIN_VARIANCE, variance );
	float d = t - moments.x;
	float pmax = linstep( VSM_LIGHT_BLEEDING_REDUCTION, 1.0, variance/(variance + d*d) );
	return max( p, pmax );
}






vec4 shading(){
	float ambientFactor = mat.lightFactors[0];
	float diffuseFactor = mat.lightFactors[1];
	float specularFactor = mat.lightFactors[3];
	float shininess = mat.lightFactors[4];
	
	vec3 ambCol = mat.ambientColor;
	float ambAlpha = 1.0;
	/*vec4 texcol = texture( mat.ambientTex,  fragmentTexCoords * mat.ambientTextureScaleUV );
	if( mat.hasAmbientTex == true ){
		ambCol = texcol.rgb;
		ambAlpha = texcol.a;
	}*/

	vec3 diffCol = mat.diffuseColor;
	float diffAlpha = 1.0;
	vec4 texcol = texture( mat.diffuseTex,  fragmentTexCoords * mat.diffuseTextureScaleUV );
	if( mat.hasDiffuseTex == true ){
		diffCol = texcol.rgb;
		diffAlpha = texcol.a;
	} 
	
	vec3 specCol = mat.specularColor;
	if( mat.hasSpecularTex == true ) specCol =  texture( mat.specularTex,  fragmentTexCoords * mat.specularTextureScaleUV ).rgb;
	
	//debug term.. set in config 
	vec3 fragcol = diffCol * debugLightFactor;
	
	//ambient term
	fragcol =  fragcol + ambCol * ambientFactor;
	
	//shadow map spot lights:
	vec2 sptDepthLightMoments[SPT_LIGHTCOUNT];
	for( int i = 0; i < SPT_LIGHTCOUNT; ++i ) sptDepthLightMoments[i] = textureProj( sptLights[i].shadowMap, fragmentLightSpt[i] ).rg;
	
	
	//Directional Light:
	for( int i = 0; i < dirLightsActive; ++i ){
		if( dirLights[i].color != vec3( 0.0, 0.0, 0.0 ) ) {

			//shadow mapping:
			float inLight = 1.0;
			/*if( useShadowMapping == true ){
				vec2 depthLightMoments = textureProj( dirLights[i].shadowMap, fragmentLightDir[i] ).rg;
				inLight = shadowVSM( depthLightMoments, fragmentLightLinearDepthDir[i] );
			}*/

			if( inLight > 0.009 ){

				vec4 Ldirection = dirLightViewSpaceDir[i];
				float cosLN = max( dot( N, -Ldirection ), 0.0 ); //because directional light comes from no point just the normalizes direction is needed to the normal vector
				fragcol += diffCol * dirLights[i].color * dirLights[i].intensity * cosLN * diffuseFactor * inLight;

				//specular part
				//PHONG
				//vec4 R = reflect( Ldirection, N );
				//float cosVCR = max( dot( VC, R ), 0.0 );
				//if( cosLdirN > 0.0 ) fragcol += specCol * pow( cosVCR, shininess ) * specularFactor * inLight;
				//BLINN:
				vec4 H = normalize( VC + (-Ldirection)); //Blinn's halfway vector
				float cosHN = max( dot( H, N ), 0.0 );
				vec3 specPart = specCol * pow( cosHN, shininess )  * specularFactor * inLight * cosLN;
				fragcol += specPart;
				if( specPart != 0.0 && diffAlpha > 0.25)  diffAlpha = smoothstep( 0.25, 1.0, diffAlpha ); //-> to make parts with specular visiable

			}
		}
	}


	//Point lights
	for( int i = 0; i < pointLightsActive; ++i ){
		if( ptLights[i].color != vec3( 0.0, 0.0, 0.0 ) ) {
			vec4 Lposition = ptLightViewSpacePos[i]; //to eye space
			float intensity = calcIntensityAfterDecay( Lposition, ptLights[i].decayType, ptLights[i].intensity );

			if( intensity > 0.009 ){
				vec4 L = normalize(Lposition - V);
				float cosLN = max( dot( N, L ), 0.0 );
				fragcol += diffCol * ptLights[i].color * intensity * cosLN * diffuseFactor;

				//specular part
				//PHONG:
				//vec4 R = reflect( -L, N );
				//float cosVCR = max( dot( VC, R ), 0.0 );
				//if( cosVCR > 0 && cosLN > 0.0 ) fragcol += specCol * pow( cosVCR, shininess ) * specularFactor;
				//BLINN:
				vec4 H = normalize( VC + L); //Blinn's halfway vector
				float cosHN = max( dot( H, N ), 0.0 );
				vec3 specPart = specCol * pow( cosHN, shininess ) * specularFactor * cosLN * intensity;
				fragcol += specPart;
				if( specPart != 0.0 && diffAlpha > 0.25)  diffAlpha = smoothstep( 0.25, 1.0, diffAlpha ); //-> to make parts with specular visiable

			}
		}
	}


	//Spot lights
	for( int i = 0; i < spotLightsActive; ++i ){
		if( sptLights[i].color != vec3( 0.0, 0.0, 0.0 ) ) {
			//diffuse part
			vec4 Lposition = sptLightViewSpacePos[i]; //to eye space
			float intensity = calcIntensityAfterDecay( Lposition, sptLights[i].decayType, sptLights[i].intensity );
			if( intensity > 0.009 ){

				//shadow mapping:
				float inLight = 1.0;
				//vec2 depthLightMoments = textureProj( sptLights[i].shadowMap, fragmentLightSpt[i] ).rg;
				vec2 depthLightMoments = sptDepthLightMoments[i];
				if( useShadowMapping == true ){	
					inLight = shadowVSM( depthLightMoments, fragmentLightLinearDepthSpt[i] );
					//if( inLight == 0.0 ) return vec3( 1.0, 0.0, 0.0 );
				}

				if( inLight > 0.008 ){

					vec4 Ldirection =  sptLightViewSpaceDir[i];
					
					vec4 L = normalize(Lposition - V);
					float cosLN = max( dot( N, L ), 0.0 );

					//check if it is in spotlight, outside or in overpass 
					float radLdirVLpos = acos(dot( Ldirection, -L));
					float comparatorInner = sptLights[i].innerConeAngle * 0.5;
					float comparatorOuter = 0.001 + sptLights[i].outerConeAngle * 0.5; //0.01 ...just because smoothstep else delivers 0 if edge0 and edge1 are the same
					//float insideOutside = clamp( (comparatorOuter - radLdirVLpos) / (comparatorOuter - comparatorInner ), 0.0, 1.0 );
					float insideOutside = smoothstep( comparatorOuter, comparatorInner,  radLdirVLpos); 

					fragcol += diffCol * sptLights[i].color * cosLN * diffuseFactor * intensity  * insideOutside * inLight; 

					//specular part
					//PHONG:
					/*vec4 R = reflect( Ldirection, N );
					float cosVCR = max( dot( VC, R ), 0.0 );
					vec3 specPart= specCol * pow( cosVCR, shininess ) * specularFactor * insideOutside * inLight;*/
					//BLINN:
					vec4 H = normalize( VC + L); //Blinn's halfway vector
					float cosHN = max( dot( H, N ), 0.0 );
					vec3 specPart = specCol * pow( cosHN, shininess )  * specularFactor * insideOutside * inLight * cosLN * intensity;
					fragcol += specPart;
					if( specPart != 0.0 && diffAlpha > 0.25)  diffAlpha = smoothstep( 0.25, 1.0, diffAlpha ); //-> to make parts with specular visiable

				}
			}
		}
	}

	return vec4(fragcol, diffAlpha * ambAlpha);
}


//FOG:
const float  fogStartDistance = 1000.0;
const float  fogMaxDistance   = 1500.0;
const vec4   fogColor		  = vec4( 0.0, 0.0, 0.0, 1.0 );

vec4 useFog( vec4 color ){
	float fogFactor = (fogStartDistance - -V.z)/(fogMaxDistance - fogStartDistance);
	fogFactor = clamp( fogFactor, 0.0, 1.0 );
	return mix(fogColor, color, fogFactor);
}


//http://www.neilmendoza.com/glsl-rotation-about-an-arbitrary-axis/
mat4 rotationMatrix(vec3 axis, float angle)
{
    axis = normalize(axis);
    float s = sin(angle);
    float c = cos(angle);
    float oc = 1.0 - c;
    
    return mat4(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,  0.0,
                oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,  0.0,
                oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c,           0.0,
                0.0,                                0.0,                                0.0,                                1.0);
}

//Calculate normal from bump map (==heightmap)!
//better not use! expensive
#define TEXCO_OFFSET 0.01
vec4 getNormalFromHeightMap( float bumpHeight )
{
	vec2 texco = fragmentTexCoords * mat.bumpTextureScaleUV;
	float xp  = texture( mat.bumpMap, vec2( texco.x+0.01, texco.y 		) ).r;
	float yp  = texture( mat.bumpMap, vec2( texco.x 	, texco.y+0.01) ).r;
	//float xm  = texture( mat.bumpMap, vec2( texco.x-0.01, texco.y 		) ).r;
	//float ym  = texture( mat.bumpMap, vec2( texco.x 	, texco.y-0.01) ).r;
	//return vec4( (xp-xm)/2.0, (yp-ym)/2.0, bumpHeight, 0.0 );
	return vec4( (bumpHeight-xp), (bumpHeight-yp), bumpHeight, 0.0 );
}




void main(void)
{
	V = fragment; //surface position in eye space
	VC = normalize( -fragment ); //direction surface -> camera
	N = fragmentNormal;
	
	//Bump Mapping  look up in normal texture:
	if( mat.bumpFactor != 0.0 ){
		vec3 bumpMapNormal = texture( mat.bumpMap, fragmentTexCoords * mat.bumpTextureScaleUV ).rgb;
		if( mat.bumpMapIsNormalMap == true ) N = vec4(( bumpMapNormal * 2.0 ) - ( 1.0 ), 0.0);
		else N = getNormalFromHeightMap( bumpMapNormal.r );
		N.x = N.x * mat.bumpFactor;
		N.y = N.y * mat.bumpFactor;
		//N = normalize(N);
		N = ( mat4(tangentToViewSpaceMat) * N );	 //->to else its in World space
	}
	
	N = normalize( N );

	vec4 fragCol = shading();
	float transparentFactor = mat.lightFactors[2];
	fragColor = vec4(fragCol.rgb, fragCol.a * transparentFactor);

	//use premultiplied alpha for multiple transparency objects behind each other, set glBlendMode( GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	//fragColor.rgb *= fragColor.a;


	//is use fog: SET VALUES!
	//fragColor = useFog(fragColor);
}

