#version 330
#extension GL_ARB_explicit_uniform_location : enable
#pragma optionNV(unroll all)

//to deactivate light set color to 0,0,0

#define SPT_LIGHTCOUNT 3  //light count per type
#define DIR_LIGHTCOUNT 2
#define PT_LIGHTCOUNT 6

#define SHADOW_BIAS_Z 0.010 //against shadow acne


const mat4 biasMatrix = mat4(0.5, 0.0, 0.0, 0.0,
                     		 0.0, 0.5, 0.0, 0.0,
                     		 0.0, 0.0, 0.5, 0.0,
                     		 0.5, 0.5, 0.5, 1.0);


layout( location = 0 ) uniform mat4	projectionMatrix; //eye space -> homogenous clip space (-> perspective division -> device coords)
layout( location = 4 ) uniform mat4	viewMatrix; //world space -> eye space
layout( location = 7 ) uniform mat4	modelMatrix;

layout( location = 11) uniform bool	instancingUsed; //-> used in mat4 modelMatrixI instead of the uniform modelMatrix!
in		mat4	modelMatrixInstance; 


struct DirectionalLight
{
	vec3  color;
	vec4  direction; //standart direciton is 0.0, 0.0, -1.0,... than it gets rotated
	float intensity;
	sampler2D shadowMap;

	//just used in vertex shader:
	mat4  orthoProjMat;
	mat4  transformMat;
	float zFarPlane;
};

struct PointLight
{
	vec3  color;
	vec4  position; //has position & rotation
	float intensity;
	int   decayType;

	//just used in vertex shader:
	mat4  persProjMat;
	mat4  transformMat;
	float zFarPlane;
};

struct SpotLight
{
	vec3  color;
	vec4  position;
	vec4  direction; //standart direciton is 0.0, 0.0, -1.0,... than it gets rotated
	float intensity;
	float innerConeAngle;
	float outerConeAngle;
	int   decayType;
	sampler2D shadowMap;

	//just used in vertex shader
	mat4  persProjMat;
	mat4  transformMat;
	float zFarPlane;
};

//Material:
struct Material
{
	float   lightFactors[5]; //ambientFactor, diffuseFactor, transparency, specularFactor, shininess  
	vec3	ambientColor; 				
	vec3 	diffuseColor; 				

	//texture
	bool	  hasAmbientTex;            
	sampler2D ambientTex;
	vec2 	  ambientTextureScaleUV;
	
	bool	  hasDiffuseTex;
	sampler2D diffuseTex;
	vec2 	  diffuseTextureScaleUV;
	
	float	  bumpFactor;    //if no bump map this is also set to 0
	bool	  bumpMapIsNormalMap; //-> normal map
	sampler2D bumpMap;
	vec2      bumpTextureScaleUV;

	vec3	  specularColor;
	bool      hasSpecularTex;
	sampler2D specularTex;
	vec2      specularTextureScaleUV;
};

layout( location = 20 ) uniform Material mat;




//for the light transformations for shadow mapping and more
uniform SpotLight			sptLights[ SPT_LIGHTCOUNT ];
uniform DirectionalLight	dirLights[ DIR_LIGHTCOUNT ];
uniform PointLight			ptLights[ PT_LIGHTCOUNT ];
layout( location = 50 ) uniform int spotLightsActive;
layout( location = 51 ) uniform int dirLightsActive;
layout( location = 52 ) uniform int pointLightsActive;


//they are set in the beginning anyway so no performance impact while running
uniform float zNearPlane;

uniform bool useShadowMapping;


in vec4 vertex; //vertexposition
in vec4 vertexNormal; //vertexnormals
in vec2 textureCoords;
in vec4 tangents; //bump mapping



//bump mapping -> transformation matrix for transforming the normal vector out of the bump map into the view space:
out mat3 tangentToViewSpaceMat;


//outs are then in world space
flat out vec4 				camWorldPos;
out vec4 					fragmentNormal;  
out vec4 					fragment;		 
out vec2 					fragmentTexCoords;

//calculate the lights in view space:
out vec4 sptLightViewSpacePos[ SPT_LIGHTCOUNT ];
out vec4 sptLightViewSpaceDir[ SPT_LIGHTCOUNT ];
out vec4 ptLightViewSpacePos[  PT_LIGHTCOUNT  ];
out vec4 dirLightViewSpaceDir[ DIR_LIGHTCOUNT ];


out vec4 fragmentLightSpt[ SPT_LIGHTCOUNT ]; //vertex positions in spot light space with perspective projection
out float fragmentLightLinearDepthSpt[ SPT_LIGHTCOUNT ]; //depth of the vertex in ligth space

//out vec4 fragmentLightDir[ DIR_LIGHTCOUNT ]; //vertex positions in directional light space
//out float fragmentLightLinearDepthDir[ DIR_LIGHTCOUNT ];

//not used yet
//out vec4 fragmentLightPt[ PT_LIGHTCOUNT ]; //vertex positions in directional light space
//out float fragmentLightLinearDepthPt[ PT_LIGHTCOUNT ];




float linstep(float low, float high, float v){
    return clamp( (v - low)/(high - low), 0.0, 1.0);
}



float normalizeDepth( float depth, float zFarPlane ){
	return linstep( zNearPlane, zFarPlane, depth );
}




void main(void)
{
	mat4 modelMatrixUsed = modelMatrix;
	if( instancingUsed == true ) modelMatrixUsed = modelMatrixInstance;
	
	vec4 modelVertex = modelMatrixUsed * vertex; 
	mat4 modelViewMatrix = viewMatrix * modelMatrixUsed;

	//extract the rotation matrix since only this is needed for the normal transformation (need only to be rotated since its normalized and not moved anywhere from the points)
	mat3 modelViewRotMatrix = mat3( modelViewMatrix[0][0], modelViewMatrix[0][1], modelViewMatrix[0][2], modelViewMatrix[1][0], modelViewMatrix[1][1], modelViewMatrix[1][2], modelViewMatrix[2][0], modelViewMatrix[2][1], modelViewMatrix[2][2] );
	mat3 normalMatrix = transpose( inverse( modelViewRotMatrix ) );

	fragmentNormal = vec4( normalize( normalMatrix * vertexNormal.xyz ), 0.0);
	
	fragment = modelViewMatrix * vertex; //just the oppisotion of the view ray, but not the projection perspective needed

	fragmentTexCoords = textureCoords;
	
	gl_Position = projectionMatrix *  fragment; //vertex transformation, right transformation happens first
	
	
	//bump mapping: -> transformation matrix for transforming the normal vector out of the bump map into the view space:
	tangentToViewSpaceMat = mat3(1.0);
	if( mat.bumpFactor != 0.0 ){
		tangentToViewSpaceMat[0] = normalize( modelViewRotMatrix * tangents.xyz );
		tangentToViewSpaceMat[2] = fragmentNormal.xyz;
		tangentToViewSpaceMat[1] = normalize( cross( fragmentNormal.xyz, tangentToViewSpaceMat[0] ) );
	}

	for( int i = 0; i < spotLightsActive; ++i ){
		sptLightViewSpacePos[i] = viewMatrix * sptLights[i].position;
		sptLightViewSpaceDir[i] = normalize(viewMatrix * sptLights[i].direction);
	}

	for( int i = 0; i < pointLightsActive; ++i ){
		ptLightViewSpacePos[i] = viewMatrix * ptLights[i].position;
	}

	for( int i = 0; i < dirLightsActive; ++i ){
		dirLightViewSpaceDir[i] = normalize(viewMatrix * dirLights[i].direction);
	}

	
	
	//SHADOW MAPPING:
	if( useShadowMapping == true ){
		
		//create this point in the light space
		for( int i = 0; i < spotLightsActive; ++i ){
			//fragmentLightSpt[i] = vec4( 0.0 );
			if( sptLights[i].zFarPlane != 0.0 ){
				fragmentLightSpt[i] = sptLights[i].transformMat * modelVertex; //in light space 
				fragmentLightLinearDepthSpt[i] = normalizeDepth( length( fragmentLightSpt[i] ), sptLights[i].zFarPlane ); //take z in the eye space of the light and normalize it within the frumstrum, - because it's negative z in opengl
				fragmentLightLinearDepthSpt[i] -= SHADOW_BIAS_Z; //agains shadow acne -> z values more distinct
				fragmentLightSpt[i] = biasMatrix * sptLights[i].persProjMat * fragmentLightSpt[i]; //in projective screen space
			}
		}


		/*for( int i = 0; i < dirLightsActive; ++i ){
			//fragmentLightDir[i] = vec4( 0.0 );
			if( dirLights[i].zFarPlane != 0.0 ){
				fragmentLightDir[i] = dirLights[i].transformMat * modelVertex; 
				fragmentLightLinearDepthDir[i] = normalizeDepth( -fragmentLightDir[i].z, dirLights[i].zFarPlane );   //take z in the eye space of the light and normalize it within the frumstrum, - because it's negative z in opengl
				fragmentLightLinearDepthDir[i] -= SHADOW_BIAS_Z; //agains shadow acne -> z values more distinct
				fragmentLightDir[i] = biasMatrix * dirLights[i].orthoProjMat * fragmentLightDir[i];
			}
		}*/

	}

}
